# Copyright (c) 2019, Digi International, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any, Collection, List, Optional

PWRON_RESET = ...  # type: int
HARD_RESET = ...  # type: int
WDT_RESET = ...  # type: int
DEEPSLEEP_RESET = ...  # type: int
SOFT_RESET = ...  # type: int
BROWNOUT_RESET = ...  # type: int
LOCKUP_RESET = ...  # type: int
CLEAN_SHUTDOWN = ...  # type: int

PIN_WAKE = ...  # type: int
RTC_WAKE = ...  # type: int


def unique_id() -> bytes:
    """
    Returns a 64-bit bytes object with a unique identifier for the XBee device.

    :return: A 64-bit bytes object with a unique identifier for the XBee
        device.
    """
    ...


def reset() -> None:
    """
    Resets the device in a manner similar to pushing the external RESET button.
    """
    ...


def soft_reset() -> None:
    """
    Performs a soft-reset the REPL.
    """
    ...


def rng() -> int:
    """
    Returns a 30-bit random number that is generated by the software.

    :return: A 30-bit random number that is generated by the software.
    """
    ...


def reset_cause() -> int:
    """
    Returns the cause of a reset. These are possible return values:

    * ``PWRON_RESET``
    * ``HARD_RESET``
    * ``WDT_RESET``
    * ``DEEPSLEEP_RESET``
    * ``SOFT_RESET``
    * ``BROWNOUT_RESET``
    * ``LOCKUP_RESET``

    :return: The reset cause.
    """
    ...


class Pin(object):
    """
    A pin is the basic object to control I/O pins. It has methods to set the
    mode of the pin (input, output, etc) and methods to get and set the digital
    logic level. For analog control of a pin, see the ``ADC`` class.
    """

    ALT = ...  # type: int
    ALT_OPEN_DRAIN = ...  # type: int
    ANALOG = ...  # type: int
    DISABLED = ...  # type: int
    IN = ...  # type: int
    OPEN_DRAIN = ...  # type: int
    OUT = ...  # type: int

    PULL_UP = ...  # type: int
    PULL_DOWN = ...  # type: int

    AF0_COMMISSION = ...  # type: int
    AF1_I2C_SCL = ...  # type: int
    AF1_SPI_ATTN = ...  # type: int
    AF2_SPI_CLK = ...  # type: int
    AF2_UART1_RTS = ...  # type: int
    AF3_SPI_SS = ...  # type: int
    AF3_UART1_CTS = ...  # type: int
    AF4_SPI_MOSI = ...  # type: int
    AF4_UART1_TX = ...  # type: int
    AF5_ASSOC_IND = ...  # type: int
    AF6_RTS = ...  # type: int
    AF7_CTS = ...  # type: int
    AF7_RS485_ENABLE_HIGH = ...  # type: int
    AF7_RS485_ENABLE_LOW = ...  # type: int
    AF8_SLEEP_REQ = ...  # type: int
    AF9_ON_SLEEP = ...  # type: int
    AF10_RSSI = ...  # type: int
    AF11_I2C_SDA = ...  # type: int
    AF11_TEMP_FAN = ...  # type: int
    AF11_USB_DIRECT = ...  # type: int
    AF12_SPI_MISO = ...  # type: int
    AF12_UART1_RX = ...  # type: int

    def __init__(self, id: Any, mode: int = -1, pull: int = -1, *,
                 value: Optional[int] = None, alt: Optional[int] = None) \
            -> None:
        """
        Class constructor. Instantiates a new ``Pin`` object with the provided
        parameters. This class represents a pin peripheral (GPIO pin)
        associated with the given ``id``.

        :param id: Object within the ``machine.Pin.board`` module or a string
            that matches one of these objects.
        :param mode: Specifies the pin mode, which can be one of:

            * ``Pin.IN`` - Pin is configured for input. If viewed as an output
              the pin is in high-impedance state.
            * ``Pin.OUT` - Pin is configured for (normal) output.
            * ``Pin.OPEN_DRAIN`` - Pin is configured for open-drain output.
              Open-drain output works in the following way: if the output value
              is set to **0** the pin is active at a low level; if the output
              value is **1** the pin is in a high-impedance state. Not all
              ports implement this mode, or some might only on certain pins.
            * ``Pin.ALT`` - Pin is configured to perform an alternative
              function, which is port specific. For a pin configured in such a
              way any other Pin methods (except ``Pin.init()``) are not
              applicable (calling them will lead to undefined, or a
              hardware-specific, result). Not all ports implement this mode.
            * ``Pin.ALT_OPEN_DRAIN`` - The Same as ``Pin.ALT``, but the pin is
              configured as open-drain. Not all ports implement this
              mode.
        :param pull: Specifies whether the pin is configured with an internal
            (weak) pull resistor attached. One of:

            * ``None`` - No pull up or down resistor.
            * ``Pin.PULL_UP`` - Pull up resistor enabled.
            * ``Pin.PULL_DOWN`` - Pull down resistor enabled.
        :param value: Valid only for ``Pin.OUT`` and ``Pin.OPEN_DRAIN`` modes
            and specifies initial output pin value if given, otherwise the
            state of the pin peripheral remains unchanged.
        :param alt: Specifies an alternate function for the pin and the values
            it can take are product dependent. This argument is valid only for
            ``Pin.ALT`` and ``Pin.ALT_OPEN_DRAIN`` modes. It may be used when a
            pin supports more than one alternate function. If only one pin
            alternate function is supported this argument is not required. Not
            all ports implement this argument.
        """
        ...

    def __call__(self, value: Optional[int] = None) -> Optional[int]:
        """
        Pin objects are callable. The call method provides a (fast) shortcut
        to set and get the value of the pin. It is equivalent to
        ``Pin.value([x])``. See ``Pin.value()`` for more details.

        :param value: Value to set on the pin.

        :return: When reading the pin, the method returns its value. When
            setting the value this method returns ``None``.
        """
        ...

    def init(self, mode: int, pull: int = -1, *,
             value: Optional[int] = None, alt: Optional[int] = None) -> None:
        """
        Re-initializes the pin using the given parameters. Only those arguments
        that are specified will be set. The rest of the pin peripheral state
        will remain unchanged. See the constructor documentation for details
        of the arguments.

        :param mode: Pin mode.
        :param pull: Flag that specifies if the pin is configured with an
            internal (weak) pull resistor attached.
        :param value: Initial output pin value.
        :param alt: Alternate function for the pin.
        """
        ...

    def value(self, value: Optional[int] = None) -> Optional[int]:
        """
        Gets or sets the value of the pin, depending on whether the argument
        ``value`` is supplied or not.

        If the argument is omitted then this method gets the digital logic
        level of the pin, returning **0** or **1** corresponding to low and
        high voltage signals respectively. The behaviour of this method depends
        on the mode of the pin:

        * ``Pin.IN``- The method returns the actual input value currently
          present on the pin.
        * ``Pin.OUT`` - The behaviour and return value of the method is
          undefined.
        * ``Pin.OPEN_DRAIN`` - If the pin is in state **0** then the behaviour
          and return value of the method is undefined. Otherwise, if the pin is
          in state **1**, the method returns the actual input value currently
          present on the pin.

        :param value: Value to set on the pin.

        :return: When reading the pin, the method returns its value. When
            setting the value this method returns ``None``.
        """
        ...

    def off(self) -> None:
        """
        Sets pin to **0** output level.
        """
        ...

    def on(self) -> None:
        """
        Sets pin to **1** output level.
        """
        ...

    def toggle(self) -> None:
        """
        Toggles the value of the pin. From **0** to **1** or from **1** to
        **0**.
        """
        ...

    def name(self) -> str:
        """
        Returns the name of the pin.

        :return: The name of the pin.
        """
        ...

    def names(self) -> List:
        """
        Returns a list with the CPU and board names for the pin.

        :return: A list with the CPU and board names for the pin.
        """
        ...

    def af_list(self) -> List:
        """
        Returns a list of alternate functions available for this pin.

        :return: A list of alternate functions available for this pin.
        """
        ...

    def mode(self, mode: Optional[int] = None) -> Optional[int]:
        """
        Gets or sets the pin mode.

        When called without a parameter, returns an integer matching one of the
        allowed constants for the mode argument to the init function. Returns
        ``None`` when setting the mode.

        :param mode: Mode to be set on a pin, which can be one of the following
            values:

            * ``Pin.IN`` - Pin is configured for input. If viewed as an output
              the pin is in high-impedance state.
            * ``Pin.OUT`` - Pin is configured for (normal) output.
            * ``Pin.OPEN_DRAIN`` - Pin is configured for open-drain output.
              Open-drain output works in the following way: if the output value
              is set to `0` the pin is active at a low level; if the output
              value is `1` the pin is in a high-impedance state. Not all ports
              implement this mode, or some might only on certain pins.
            * ``Pin.ALT`` - Pin is configured to perform an alternative
              function, which is port specific. For a pin configured in such a
              way any other Pin methods (except ``Pin.init()``) are not
              applicable (calling them will lead to undefined, or a
              hardware-specific, result). Not all ports implement this mode.
            * ``Pin.ALT_OPEN_DRAIN`` - The Same as ``Pin.ALT``, but the pin is
              configured as open-drain. Not all ports implement this mode.

        :return: Current mode on the pin or ``None`` when setting the mode.
        """
        ...

    def pull(self, pull: Optional[int] = None) -> Optional[int]:
        """
        Gets or sets the pin pull state.

        When called without a parameter, returns an integer matching one of the
        allowed constants for the mode argument to the init function. Returns
        ``None`` when setting the pull.

        :param pull: Specifies if the pin has a (weak) pull resistor attached,
            and can be one of:

            * ``None`` - No pull up or down resistor.
            * ``Pin.PULL_UP`` - Pull up resistor enabled.
            * ``Pin.PULL_DOWN`` - Pull down resistor enabled.

        :return: Current pin pull state or ``None`` when setting the pull.
        """
        ...

    def af(self) -> int:
        """
        Returns the currently configured alternate-function of the pin. The
        integer returned will match one of the allowed constants for the af
        argument to the init function.

        :return: The currently configured alternate-function of the pin.
        """
        ...

    class board:
        """
        The board class contains the list of pre-defined pins.
        """

        D0 = ...  # type: Pin
        D1 = ...  # type: Pin
        D2 = ...  # type: Pin
        D3 = ...  # type: Pin
        D4 = ...  # type: Pin
        D5 = ...  # type: Pin
        D6 = ...  # type: Pin
        D7 = ...  # type: Pin
        D8 = ...  # type: Pin
        D9 = ...  # type: Pin
        D10 = ...  # type: Pin
        D11 = ...  # type: Pin
        D12 = ...  # type: Pin
        P0 = ...  # type: Pin
        P1 = ...  # type: Pin
        P2 = ...  # type: Pin


class ADC(object):
    """
    This class represents an ADC (Analog to Digital Conversion) associated to
    a physical pin on the XBee device. It can be used to read analog values
    from sensors connected to that pin.

    The ADC reading value has a resolution of 12 bits with a range of 0 - 4095.
    """

    def __init__(self, id: str) -> None:
        """
        Class constructor. Instantiates a new ``ADC`` object associated with
        the given pin.

        :param id: Pin identifier for the ADC. String that matches one of the
            ``Pin.board`` objects.
        """
        ...

    def read(self) -> int:
        """
        Reads and returns a raw ADC sample with a range of 12 bits
        (0 to 4095).

        Use the following equation to convert this value to mV::

            sample mV = (A/D reading * Vref mV) / 4095

        :return: A raw ADC sample in 12-bit range
        """
        ...

    def read_u16(self) -> int:
        """
        Reads and returns a raw ADC sample with a range of 16 bits
        (0 to 65535). This method is provided for compatibility with
        other MicroPython implementations.

        Since XBee devices only support 12-bit ADC readings,
        readings from ``read_u16()`` will match those from ``read()``,
        only scaled to a 16-bit range.

        Use the following equation to convert this value to mV::

            sample mV = (A/D reading * Vref mV) / 65535

        **Note:**

        The ``read_u16`` method is available on XBee Cellular and
        XBee 3 Cellular products with version ending in 16 and newer,
        and XBee 3 RF products with version ending in 0B and newer.

        :return: A raw ADC sample in 16-bit range
        """
        ...


class PWM(object):
    """
    This class represents a PWM (Pulse Width Modulation) associated to a
    physical pin on the XBee device. PWM generates a square wave (signal
    switched between on and off) that can be used to control the average
    power delivered to a load. The square wave is configured with the
    **frequency** and **duty cycle** parameters.

    **Note**: It is not possible to change fixed frequency on XBee's ``P0``
    and ``P1`` pins.

    Most of XBee devices have PWM functionality on pins ``P0`` and ``P1``. See
    the documentation of your XBee device for more information about PWM
    support.
    """

    def __init__(self, pin: str) -> None:
        """
        Class constructor. Instantiates a new ``PWM`` object associated with
        the given pin.

        :param pin: Pin identifier for the PWM. String that matches one of the
            ``Pin.board`` objects.
        """
        ...

    def init(self) -> None:
        """
        Turns on PWM functionality on the associated pin.
        """
        ...

    def deinit(self) -> None:
        """
        Turns off PWM functionality on the associated pin.
        """
        ...

    def freq(self, freq: Optional[int] = None) -> Optional[int]:
        """
        Sets or gets the PWM frequency. Frequency determines how fast the PWM
        completes a cycle and therefore how fast it switches between high and
        low states.

        **Note**: It is not possible to change fixed frequency on XBee's ``P0``
        and ``P1`` pins.

        :param freq: The new frequency for the PWM.

        :return: Current frequency on the PWM or ``None`` when setting the
            frequency.
        """
        ...

    def duty(self, duty: Optional[int] = None) -> Optional[int]:
        """
        Sets or gets the PWM duty cycle. Duty cycle describes the amount of
        time the signal is in a high (on) state as a percentage of the total
        time of it takes to complete one cycle

        The duty cycle is between 0 (0%) and 1023 (100%), inclusive of the
        end points.

        :param duty: The new duty cycle for the PWM.

        :return: Current duty cycle on the PWM or ``None`` when setting the
            duty cycle.
        """
        ...


class I2C(object):
    """
    This class represents an I2C object associated with a specific two wire
    bus. I2C is a two-wire protocol for communicating between devices. At the
    physical level it consists of two wires: SCL and SDA, the clock and data
    lines respectively.

    I2C objects can be initialized when created, or initialized later on.
    Printing the I2C object gives you information about its configuration.

    The XBee device can function as an I2C master controlled by MicroPython.
    This allows you to perform basic sensing and actuation with I2C devices
    such as sensors and actuators via MicroPython without an additional
    micro-controller.
    """

    def __init__(self, id: int, *, freq: int = 400000) -> None:
        """
        Class constructor. Instantiates a new ``I2C`` object with the provided
        parameters.

        :param id: Identifies a particular I2C peripheral. This version of
            MicroPython supports a single peripheral with id **1** using
            ``DIO1`` for SCL and ``DIO11`` for SDA.
        :param freq: Maximum frequency for SCL.
        """
        ...

    def init(self, scl: Pin, sda: Pin, *, freq: int = 400000) -> None:
        """
        Initializes the I2C bus with the given arguments.

        :param scl: Pin object specifying the pin to use for SCL.
        :param sda: Pin object specifying the pin to use for SDA.
        :param freq: Maximum frequency for SCL.
        """
        ...

    def scan(self) -> Collection[int]:
        """
        Scans all I2C addresses between **0x08** and **0x77** inclusive and
        returns a list of those that respond. A device responds if it pulls the
        SDA line low after its address (including a write bit) is sent on the
        bus.
        """
        ...

    def start(self) -> None:
        """
        Generates a **START** condition on the bus (SDA transitions to low
        while SCL is high).
        """
        ...

    def stop(self) -> None:
        """
        Generates a **STOP** condition on the bus (SDA transitions to high
        while SCL is high).
        """
        ...

    def readinto(self, buf: bytearray, nack: bool = True) -> None:
        """
        Reads bytes from the bus and stores them into ``buf``. The number of
        bytes read is the length of ``buf``. An **ACK** will be sent on the bus
        after receiving all but the last byte. After the last byte is received,
        if ``nack`` is true then a **NACK** will be sent, otherwise an **ACK**
        will be sent (and in this case the slave assumes more bytes are going
        to be read in a later call).

        :param buf: Buffer to read bytes into.
        :param nack: If ``True``, then NACK will be sent after reading last
            bytes.
        """
        ...

    def write(self, buf: bytearray) -> int:
        """
        Writes the bytes from ``buf`` to the bus. Checks that an **ACK** is
        received after each byte and stops transmitting the remaining bytes if
        a **NACK** is received.

        :param buf: Buffer to write bytes from.

        :return: The number of ACKs that were received.
        """
        ...

    def readfrom(self, addr: int, nbytes: int, stop: bool = True) -> bytes:
        """
        Reads ``nbytes`` from the slave specified by ``addr``.

        :param addr: Address of slave device.
        :param nbytes: Maximum amount of bytes to be read.
        :param stop: If ``True``, then **STOP** condition is generated at the
            end of the transfer.

        :return: Data that has been read.
        """
        ...

    def readfrom_into(self, addr: int, buf: bytearray,
                      stop: bool = True) -> None:
        """
        Reads into ``buf`` from the slave specified by ``addr``. The number of
        bytes read will be the length of buf. If ``stop`` is true then a
        **STOP** condition is generated at the end of the transfer.

        :param addr: Address of slave device.
        :param buf: Buffer for storing read data.
        :param stop: If ``True``, then **STOP** condition is generated at the
            end of the transfer.
        """
        ...

    def writeto(self, addr: int, buf: bytearray,
                stop: bool = True) -> int:
        """
        Writes the bytes from ``buf`` to the slave specified by ``addr``. If a
        NACK is received following the write of a byte from ``buf`` then the
        remaining bytes are not sent. If ``stop`` is true then a **STOP**
        condition is generated at the end of the transfer, even if a **NACK**
        is received.

        :param addr: Address of slave device.
        :param buf: Buffer to write data from.
        :param stop: If ``True``, then **STOP** condition is generated at the
            end of the transfer.

        :return: The number of ACKs that were received.
        """
        ...

    def readfrom_mem(self, addr: int, memaddr: int, *,
                     addrsize: int = 8) -> bytes:
        """
        Reads ``nbytes`` from the slave specified by ``addr`` starting from the
        memory address specified by ``memaddr``. The argument ``addrsize``
        specifies the address size in bits.

        :param addr: Address of slave device.
        :param memaddr: Memory address location on a slave device to read from.
        :param addrsize: Address size in bit

        :return: Data that has been read.
        """
        ...

    def readfrom_mem_into(self, addr: int, memaddr: int, buf: bytearray, *,
                          addrsize: int = 8) -> None:
        """
        Reads into ``buf`` from the slave specified by ``addr`` starting from
        the memory address specified by ``memaddr``. The number of bytes read
        is the length of buf. The argument ``addrsize`` specifies the address
        size in bits.

        :param addr: Address of slave device.
        :param memaddr: Memory address location on a slave device to write
            into.
        :param buf: Buffer to store read data.
        :param addrsize: Address size in bits.
        """
        ...

    def writeto_mem(self, addr: int, memaddr: int, buf: bytearray, *,
                    addrsize: int = 8) -> None:
        """
        Writes ``buf`` to the slave specified by ``addr`` starting from the
        memory address specified by ``memaddr``. The argument ``addrsize``
        specifies the address size in bits.

        :param addr: Address of slave device.
        :param memaddr: Memory address location on a slave device to write
            into.
        :param buf: Buffer containing data to write.
        :param addrsize: Address size in bits.
        """
        ...


class UART(object):
    """
    This class represents a UART (Universal Asynchronous Receiver/Transmitter)
    port on the XBee device that can be used to transmit and receive serial
    data.

    **Note**: This class only applies to XBee modules that support the
    **Secondary UART** feature such as the XBee Cellular modules. See
    the documentation of your XBee device for more information about UART
    support.
    """

    RTS: ...  # type: int
    CTS: ...  # type: int

    def __init__(self, id: int,  baudrate: int, bits: int = 8,
                 parity: Optional[int] = None, stop: int = 1, *, flow: int = 0,
                 timeout: int = 0, timeout_char: int = 0) -> None:
        """
        Class constructor. Instantiates a new ``UART`` object with the given
        parameters.

        :param id: ID of UART object. **1** for those XBee modules that support
            the **Secondary UART** feature.
        :param bits: Bits per character, a value from 5 to 8.
        :param parity: An additional parity bit added to each byte, either
            ``None``, **0** (even) or **1** (odd).
        :param stop: Number of stop bits after the optional parity bit, either
            **1** or **2**.
        :param flow: Hardware flow control; either **0** for none, ``UART.RTS``
            for RTS-only, ``UART.CTS`` for CTS-only or ``UART.RTS|UART.CTS``
            for both.
        :param timeout: Number of milliseconds to wait for reading the first
            character.
        :param timeout_char: Number of milliseconds to wait between characters
            when reading.
        """
        ...

    def init(self, baudrate: int, bits: int = 8, parity: Optional[int] = None,
             stop: int = 1, *, flow: int = 0, timeout: int = 0,
             timeout_char: int = 0) -> None:
        """
        Reconfigures the UART with the given parameters.

        :param baudrate: Baud rate, that specifies how fast data is sent over
            serial line.
        :param bits: Bits per character, a value from 5 to 8.
        :param parity: An additional parity bit added to each byte, either
            ``None``, **0** (even) or **1** (odd).
        :param stop: Number of stop bits after the optional parity bit, either
            **1** or **2**.
        :param flow: Hardware flow control; either **0** for none, ``UART.RTS``
            for RTS-only, ``UART.CTS`` for CTS-only or ``UART.RTS|UART.CTS``
            for both.
        :param timeout: Number of milliseconds to wait for reading the first
            character.
        :param timeout_char: Number of milliseconds to wait between characters
            when reading.
        """
        ...

    def deinit(self) -> None:
        """
        Turns off the UART bus. After calling ``deinit()``, attempts to write
        to the UART result in an ``OSError(EPERM)`` exception but reads
        continue to pull buffered bytes.
        """
        ...

    def any(self) -> int:
        """
        Returns an integer counting the number of characters in the read
        buffer, or **0** if no bytes are available.

        :return: Number of characters in the read buffer, or **0** if no bytes
            are available.
        """
        ...

    def read(self, nbytes: Optional[int] = None) -> Optional[bytes]:
        """
        Reads characters. If ``nbytes`` is specified then reads at most that
        many bytes, otherwise reads as much data as possible.

        :param nbytes: Upper limit on number of read characters.

        :return: Bytes read in or ``None`` on timeout.
        """
        ...

    def readinto(self, buf: bytearray, nbytes: Optional[int] = None) \
            -> Optional[int]:
        """
        Reads bytes into the ``buf``. If ``nbytes`` is specified then reads at
        most that many bytes. Otherwise, reads at most ``len(buf)`` bytes.

        :param buf: Buffer for holding read data.
        :param nbytes: Upper limit on number of read characters.

        :return: Number of bytes read in or ``None`` on timeout.
        """
        ...

    def readline(self) -> Optional[bytes]:
        """
        Reads a line, ending in a newline character.

        :return: The line read or ``None`` on timeout.
        """
        ...

    def write(self, buf: bytearray) -> Optional[int]:
        """
        Writes the buffer of bytes to the bus.

        :param buf: Data that needs to be written.

        :return: Number of bytes written or ``None`` on timeout.
        """
        ...

class WDT:
    """Restarts the application when it does not indicate proper operation.

    Once started it cannot be stopped or reconfigured in any
    way. After enabling, the application must "feed" the watchdog
    periodically to prevent it from expiring and resetting the
    system.

    **Note**: Added and only supported as of the following product versions:
      * XBee3 Cellular LTE-M/NB-IoT: version 11415
      * XBee3 Cellular LTE Cat 1: version x15
    """

    def __init__(self,
                    id: int = 0,
                    timeout: int = 60000,
                    response: int = SOFT_RESET) -> None:
        """Create a WDT object and start it.

        :param id: Must be zero

        :param timeout: Time in milliseconds in which the watchdog
        must be fed before it will fire.

        :param response: Choices include:
            * ``SOFT_RESET``: Reset only the MicroPython interpretor.
            * ``HARD_RESET``: Reset the entire XBee.
            * ``CLEAN_SHUTDOWN``: On the XBee3 Cellular products,
                perform a clean shutdown of the cellular component and
                then reset the entire XBee.

        """
        ...

    def feed() -> None:
        """Resets the expiration timer to indicate proper operation."""
        ...



"""
Heap queue algorithm.

MicroPython module: https://docs.micropython.org/en/v1.25.0/library/heapq.html

CPython module: :mod:`python:heapq` https://docs.python.org/3/library/heapq.html .

This module implements the
`min heap queue algorithm <https://en.wikipedia.org/wiki/Heap_%28data_structure%29>`_.

A heap queue is essentially a list that has its elements stored in such a way
that the first item of the list is always the smallest.

---
Module: 'heapq' on micropython-v1.25.0-preview-rp2-UNKNOWN_BOARD
"""

# MCU: {'family': 'micropython', 'version': '1.25.0-preview', 'build': 'preview.236.gbfb1bee6f', 'ver': '1.25.0-preview-preview.236.gbfb1bee6f', 'port': 'rp2', 'board': 'UNKNOWN_BOARD', 'cpu': 'RP2350', 'mpy': 'v6.3', 'arch': 'armv7emsp'}
# Stubber: v1.24.0
from __future__ import annotations
from _typeshed import Incomplete
from typing import Any
from typing_extensions import Awaitable, TypeAlias, TypeVar

_T = TypeVar("_T")

def heappop(heap: list[_T], /) -> _T:
    """
    Pop the first item from the ``heap``, and return it.  Raise ``IndexError`` if
    ``heap`` is empty.

    The returned item will be the smallest item in the ``heap``.
    """
    ...

def heappush(heap: list[_T], item: _T, /) -> None:
    """
    Push the ``item`` onto the ``heap``.
    """
    ...

def heapify(x: list[Any], /) -> None:
    """
    Convert the list ``x`` into a heap.  This is an in-place operation.
    """
    ...



# Copyright (c) 2019, Digi International, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import uio

from typing import Any, Dict, List, Optional, Tuple


argv: List[str] = ...
byteorder: str = ...
implementation: Tuple[str, Tuple[int, int, int], int] = ...
maxsize: int = ...
modules: Dict[str, Any] = ...  # technically [str, Module]
path: List[str] = ...
platform: str = ...
stderr: uio.FileIO = ...
stdin: uio.FileIO = ...
stdout: uio.FileIO = ...
version: str = ...
version_info: Tuple[int, int, int] = ...


def exit(arg: Optional[object] = None, /) -> None:
    """
    Raise a ``SystemExit`` exception, with the given argument if specified.

    Note that calling ``sys.exit()`` while at the MicroPython REPL
    (``>>>`` prompt) has no effect.
    """


def print_exception(exc: BaseException, file: uio._IOBase = stdout, /) -> None:
    """
    Print the given exception and its traceback to a file-like object
    (default is ``sys.stdout``).

    This is a simplified version of CPython's ``traceback.print_exception()``
    function.
    """




import utime
import ustruct
import machine

from micropython import const
import xbee 

def get_cpu_temp() -> int:    
    """
    Read the internal CPU temperature
    
    :return: The cpu temperature value 
    """
            
class Led():
    """
    The Led object is used to control the state of the LED.
    """
    
    def on(self) -> None:
        """
        light on the LED.
        """
        
    def off(self) -> None:
        """
        light off the LED.
        """
 
    def toggle(self) -> None:
        """
        Toggle the LED state.
        """
    
    def state(self) -> bool:
        """
        Return the state of the LED.
        
        :return: ``True`` if the LED is on, ``False`` if the LED is off.
        """
 

class SupplyVoltage(machine.ADC):
    """
    The supply voltage sensor is an analog sensor that measures the supply voltage.
    """

    def read(self) -> float:
        """
        Reads the supply voltage.
        
        :return: The supply voltage.
        """


class Illuminance:
    """
    The illuminance sensor is a digital sensor that measures the illuminance in lux.
    """
    
    CONTINUOUS_MODE = const(0x10)
    ONE_TIME_MODE = const(0x20)
    
    def __init__(self, scale_factor=2.8):
        """
        Initializes the illuminance sensor.
        
        :param scale_factor: The scale factor.
        """ 
        
    def read(self, continuous=True) -> int:            
        """
        Reads the illuminance value.
        
        :param continuous: If ``True`` the sensor is in continuous mode.
        
        :return: The illuminance value.
        """

class Tphg:
    """
    This object is used to read temperature, pressure, humidity and gas values.
    or to calculate the altitude and sea level pressure.
    and to calculate the IAQ index.
    """
    
    def __init__(self, temp_weighting:float=0.10, pressure_weighting:float=0.05, humi_weighting:float=0.20, gas_weighting:float=0.65, gas_ema_alpha:float=0.1, temp_baseline:float=23.0, pressure_baseline:float=1013.25, humi_baseline:float=45.0, gas_baseline:int=450_000) :
        """
        Initializes the Tphg sensor.        
              
        :param temp_weighting: Temperature weighting.
        :param pressure_weighting: Pressure weighting.        
        :param humi_weighting: Humidity weighting.
        :param gas_weighting: Gas weighting.
        :param gas_ema_alpha: Gas EMA alpha.
        :param temp_baseline: Temperature baseline.
        :param pressure_baseline: Pressure baseline.
        :param humi_baseline: Humidity baseline.
        :param gas_baseline: Gas baseline.
        """ 
        
    def set_temperature_correction(self, value:float) -> None:
        """
        Compensates for temperature.
        
        :param value: Temperature compensation value
        """


    def read(self, gas: bool=False) -> tuple:
        """
        Reads the temperature, pressure, humidity and gas values.
        
        :param gas: If ``True`` reads the gas value.
        
        :return: A tuple with temperature, pressure, and humidity values. However, if the parameter gas is True, gas is added.    
        """

        
    def sealevel(self, altitude:float) -> float:
        """
        calculates the pressure at sea level based on the altitude
        
        :param altitude: Altitude in meters.
        
        :return: The pressure at sea
        """
        
    def altitude(self, sealevel:float) -> float: 
        """
        calclates the altitude based on the sealevel pressure
        
        :param sealevel: Pressure at sea level.
        
        :return: The altitude in meters.
        """

    def iaq(self) -> tuple:
        """
        Reads the IAQ index.
        
        :return: Tuple with the IAQ index, temperature, pressure, humidity and gas.
        """

    def burnIn(self, threshold: float=0.01, count: int=10, timeout_sec: int=180) -> tuple:
        """
        Performs stabilization operations required for gas measurements.
        
        :param threshold: The threshold value.
        :param count: The number of measurements.
        :param timeout_sec: The timeout in seconds.
        
        :return: Tuple with the state, gas, deviation.
        """


# Copyright (c) 2019, Digi International, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any, List


def open(file: str, *, mode: str="r", buffering: int =-1,
         encoding: str=None) -> Any:
    """
    Opens a file. Builtin ``open()`` function is an alias to
    ``uio.open(file, mode='r')`` which returns a file object a ``uio.FileIO``
    object for binary modes and a ``uio.TextIOWrapper`` object for text modes.
    If the file cannot be opened, an ``OSError`` is raised.

    :param file: Path-like object giving the path absolute or relative to the
        current working directory of the file to be opened.
    :param mode: Optional string that specifies the mode in which the file is
        opened. It defaults to **'r'** which means open for reading in text
        mode. Other common values are **'w'** for writing (truncating the file
        if it already exists), **'x'** for exclusive creation, **'a'** for
        appending all writes append to the end of the file regardless of the
        current seek position and **'*'** for opening a secure file for writing
        (only available on modules that support secure files).
    :param buffering: Optional integer used to set the buffering policy. Pass
        **0** to switch buffering off (only allowed in binary mode), **1** to
        select line buffering (only usable in text mode), and an
        **integer > 1** to indicate the size in bytes of a fixed-size chunk
        buffer.
    :param encoding: The name of the encoding used to decode or encode the
        file. This should only be used in text mode.

    :return: A file object.
    """
    ...

class _IOBase(object):
    """
    The abstract base class for all I/O classes, acting on streams of bytes.
    There is no public constructor.
    """

    def read(self, size: int=-1) -> bytes:
        """
        Reads up to ``size`` bytes from the object and returns them. As a
        convenience, if ``size`` is unspecified or ``-1``, all bytes until
        end-of-file (EOF) are returned.

        :param size: Maximum number of bytes to read.

        :return: Read bytes.
        """
        ...

    def readinto(self, b: Any) -> int:
        """
        Reads bytes into a pre-allocated, writable bytes-like object ``b``,
        and returns the number of bytes read.

        :param b: Writable bytes-like object where read bytes will be placed.

        :return: Number of bytes read.
        """
        ...

    def readline(self, size: int=-1) -> bytes:
        """
        Reads and returns one line from the stream. If ``size`` is specified,
        at most ``size`` bytes are read.

        :param size: Maximum number of bytes to read.

        :return: Read line in bytes.
        """
        ...

    def write(self, b: Any) -> int:
        """
        Writes the given bytes-like object, ``b``, to the underlying raw
        stream, and returns the number of bytes written.

        :param b: Bytes-like object to write in the stream.

        :return: The number of bytes written.
        """
        ...

    def seek(self, offset: int, whence: int=0) -> int:
        """
        Changes the stream position to the given byte ``offset``. ``offset``
        is interpreted relative to the position indicated by ``whence``. The
        default value for ``whence`` is ``0``. Values for whence are:

        * ``0`` – start of the stream (the default); offset should be zero or
          positive.
        * ``1`` – current stream position; offset may be negative.
        * ``2`` – end of the stream; offset is usually negative.

        **Note**: Seeking is disabled when writing to secure files.

        :param offset: New stream position relative to the position indicated
            by ``whence``.
        :param whence: Initial position of the stream.

        :return: The new absolute position of the stream.
        """
        ...

    def tell(self) -> int:
        """
        Returns the current stream position.

        :return: The current stream position.
        """
        ...

    def flush(self) -> None:
        """
        Flushes the write buffers of the stream if applicable. This does
        nothing for read-only streams.
        """
        ...

    def close(self) -> None:
        """
        Flushes and closes the stream. This does nothing if the file is
        already closed.
        """
        ...

    def __enter__(self) -> Any:
        """
        """
        ...

    def __exit__(self) -> None:
        """
        """
        ...

class FileIO(_IOBase):
    """
    Represents an OS-level file containing bytes data.
    """

    def __init__(self, file: str, *, mode: str="r", buffering: int =-1,
                 encoding: str=None) -> None:
        """
        Class constructor. Instantiates a ``FileIO`` object with the provided
        parameters.

        :param file: The file name.
        :param mode: Optional string that specifies the mode in which the file
            is opened. It defaults to **'r'** which means open for reading in
            text mode. Other common values are **'w'** for writing (truncating
            the file if it already exists), **'x'** for exclusive creation,
            **'a'** for appending all writes append to the end of the file
            regardless of the current seek position and **'*'** for opening a
            secure file for writing (only available on modules that support
            secure files).
        :param buffering: Optional integer used to set the buffering policy.
            Pass **0** to switch buffering off (only allowed in binary mode),
            **1** to select line buffering (only usable in text mode), and an
            **integer > 1** to indicate the size in bytes of a fixed-size chunk
            buffer.
        :param encoding: Gives the name of the encoding that the stream will be
            decoded or encoded with.
        """
        ...

    def readlines(self) -> List:
        """
        Reads and returns a list of lines from the stream.

        **Note**: It is already possible to iterate on file objects using
            ``for line in file: ...`` without calling ``file.readlines()``.

        :return: A list of lines from the stream.
        """
        ...

    def __del__(self) -> None:
        """
        Prepares for object destruction. Calls the instance's ``close()``
        method.
        """
        ...

class TextIOWrapper(_IOBase):
    """
    Buffered text stream that provides a character and line based interface to
    stream I/O.
    """

    def __init__(self, file: str, *, mode: str="r", buffering: int =-1,
                 encoding: str=None) -> None:
        """
        Class constructor. Instantiates a ``TextIOWrapper`` with the provided
        parameters.

        :param file: Name of the file to stream.
        :param mode: Optional string that specifies the mode in which the file
            is opened. It defaults to **'r'** which means open for reading in
            text mode. Other common values are **'w'** for writing (truncating
            the file if it already exists), **'x'** for exclusive creation,
            **'a'** for appending all writes append to the end of the file
            regardless of the current seek position and **'*'** for opening a
            secure file for writing (only available on modules that support
            secure files).
        :param buffering: Optional integer used to set the buffering policy.
            Pass **0** to switch buffering off (only allowed in binary mode),
            **1** to select line buffering (only usable in text mode), and an
            **integer > 1** to indicate the size in bytes of a fixed-size chunk
            buffer.
        :param encoding: Gives the name of the encoding that the stream will be
            decoded or encoded with.
        """
        ...

    def readlines(self) -> List:
        """
        Reads and returns a list of lines from the stream.

        **Note**: It is already possible to iterate on file objects using
            ``for line in file: ...`` without calling ``file.readlines()``.

        :return: A list of lines from the stream.
        """
        ...

    def __del__(self) -> None:
        """
        Prepares for object destruction. Calls the instance's ``close()``
        method.
        """
        ...

class StringIO(_IOBase):
    """
    An in-memory stream for text I/O. The text buffer is discarded when the
    ``close()`` method is called.
    """

    def __init__(self, *, initial_value: str="", newline: str="\n") -> None:
        """
        Class constructor. Instantiates a ``StringIO`` object with the provided
        parameters.

        The initial value of the buffer can be set by providing
        ``initial_value``. If newline translation is enabled, newlines
        will be encoded as if by ``write()``. The stream is positioned at the
        start of the buffer.

        :param initial_value: String with the initial value of the buffer.
        :param newline: Controls how line endings are handled. It can be
            ``None``, ``''``, ``'\n'``, ``'\r'``, and ``'\r\n'``.
        """
        ...

    def getvalue(self) -> str:
        """
        Returns a string containing the entire contents of the buffer.
        Newlines are decoded as if by ``read()``, although the stream position
        is not changed.

        :return: String with the contents of the buffer.
        """
        ...

class BytesIO(_IOBase):
    """
    A stream implementation using an in-memory bytes buffer. The buffer is
    discarded when the ``close()`` method is called.
    """

    def __init__(self, initial_bytes: Any=None) -> None:
        """
        Class constructor. Instantiates a ``BytesIO`` object with an optional
        initial data.

        :param initial_bytes: bytes-like-object with initial data.
        """
        ...

    def getvalue(self) -> bytes:
        """
        Gets the current contents of the underlying buffer which holds data.

        :return: Current contents of the underlying buffer in bytes format.
        """
        ...



# Copyright (c) 2020, Digi International, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any, Optional, Callable

try:
    # Python 3.8 (and PyCharm 2019.2.5 and newer) support TypedDict,
    # which means the _GAPScanDict type can actually provide useful hints.
    # (PyCharm issue PY-36008.)
    # In case the PyCharm installation being used does not have TypedDict defined,
    # fall back to using just a regular dict type.

    from typing import TypedDict
    class _RelayDict(TypedDict):
        message: str
        sender: str
except ImportError:
    from typing import Dict
    _RelayDict = Dict


SERIAL: int = ...
BLUETOOTH: int = ...
MICROPYTHON: int = ...
MAX_DATA_LENGTH: int = ...

def callback(callback: Optional[Callable[[_RelayDict], Any]], /) -> None:
    """
    Register a callback method that is called whenever a relay frame is
    received.

    This callback takes one parameter, a dictionary with the
    following keys:

        * ``sender``: The source interface. One of:

            * ``relay.SERIAL``
            * ``relay.BLUETOOTH``
            * ``relay.MICROPYTHON``
        * ``message``: The received data in bytearray format.

    **Note: This is only available on XBee 3 Cellular products with
    ending in 15 or newer.**

    :param callback: A function that is called whenever a relay frame is
        received.
        If ``callback`` is None, the registered callback will be
        unregistered, allowing for polling of relay.receive() again.
    """

def receive() -> _RelayDict:
    """
    Returns a relay data entry, a dictionary containing the
    following elements:

    * ``sender``: The source interface. One of:

        * ``relay.SERIAL``
        * ``relay.BLUETOOTH``
        * ``relay.MICROPYTHON``
    * ``message``: The received data in bytearray format.

    :return: A received relay data entry as a dictionary.
    """
    ...

def send(dest: int, data: Any) -> None:
    """
    Sends data to a destination interface. You can send a maximum of
    ``relay.MAX_DATA_LENGTH`` bytes in a single frame.

    The destination interface to transmit the data is one of the following:

        * ``relay.SERIAL``
        * ``relay.BLUETOOTH``
        * ``relay.MICROPYTHON`` (for loopback)

    The ``send()`` method throws exceptions in at least the following cases:

    * ``ValueError`` or ``TypeError`` for invalid parameters.
    * ``OSError(ENOBUFS)`` if unable to allocate a buffer for the frame.
    * ``OSError(ENODEV)`` for invalid ``dest`` parameter.
    * ``OSError(ECONNREFUSED)`` when destination is not accepting frames (for
      example, the serial interface is not in API mode, Bluetooth is not
      connected and unlocked, the queue is full or delivery failed).

    :param dest: Destination interface to transmit the data.
    :param data: Data to send in bytearray or string format or any other object
        that implements the buffer protocol.
    """
    ...
